import logging as lg

from easysam.utils import get_aws_client


def validate(cliparams: dict, resources_data: dict, environment: str, errors: list[str]):
    '''
    Validate required external cloud resources.

    Args:
        cliparams (dict): The CLI parameters (used: aws_profile)
        resources_data (dict): The resources data.
        environment (str): The environment name.
        errors (list[str]): The list of errors.
    '''

    iam = get_aws_client('iam', cliparams)
    ssm = get_aws_client('ssm', cliparams)
    lambdas = get_aws_client('lambda', cliparams)

    validate_bucket_policy(iam, resources_data, environment, errors)
    validate_custom_layers(ssm, lambdas, resources_data, errors)


def validate_bucket_policy(iam, resources_data, environment, errors):
    for bucket, details in resources_data.get('buckets', {}).items():
        if policy_name := details.get('extaccesspolicy'):
            full_policy_name = f'{policy_name}-{environment}'
            lg.info(f"Validating bucket policy: {full_policy_name}")

            try:
                paginator = iam.get_paginator('list_policies')
                policy = None

                for page in paginator.paginate(Scope='Local'):
                    policies = page['Policies']

                    for item in policies:
                        lg.info(f"Found policy: {item['PolicyName']}")

                        if item['PolicyName'] == full_policy_name:
                            policy = item
                            break

            except Exception as e:
                lg.error(f"Error listing policy {policy_name}: {e}")
                policy = None

            if not policy:
                errors.append(
                    f"Bucket '{bucket}' has an invalid extaccesspolicy: {policy_name}. "
                    f"Please create a policy with the name {full_policy_name}."
                )


def validate_custom_layers(ssm, lambdas, resources_data, errors):
    for function, details in resources_data.get('functions', {}).items():
        layers = details.get('layers', {})

        for layer, layer_handle in layers.items():
            if layer_handle.startswith('{{resolve:'):
                param_uri = layer_handle.split('resolve:')[1].split('}}')[0]

                if not param_uri.startswith('ssm:'):
                    errors.append(
                        f'Custom layer {layer} by URI in ({function}) is not yet supported'
                    )
                    continue

                ssm_param = param_uri.split('ssm:')[1]
                ssm_param = ssm_param.lstrip('/')
                lg.info(f'Validating SSM layer name: {ssm_param}')

                try:
                    layer_handle = ssm.get_parameter(Name=ssm_param)['Parameter']['Value']
                    lg.info(f'Successfully resolved SSM layer name: {ssm_param}')
                    lg.info(f'Layer ARN received: {layer_handle}')

                except Exception:
                    errors.append(f'SSM parameter {ssm_param} not found')
                    continue

            if layer_handle.startswith('arn:'):
                lg.info(f'Looking for layer ARN: {layer_handle}')

                try:
                    response = lambdas.get_layer_version_by_arn(Arn=layer_handle)
                    version = response['Version']
                    lg.info(f'Layer ARN found: {layer_handle}, version: {version}')
                    continue

                except Exception:
                    errors.append(f'Layer ARN {layer_handle} not found')
                    continue

            errors.append(f"Custom layer {layer} in function {function} is not supported")



import logging as lg
from pathlib import Path
import ast


def commondep(common_base, target_dir):
    common_base = Path(common_base)
    target_dir = Path(target_dir)
    commons = find_commons(common_base)

    lg.debug(f'Commons: {commons}')
    common_imports = find_common_deps(target_dir, common_base, commons, set())
    return sorted(list(common_imports))


def find_commons(common_base):
    commons = []

    for item in common_base.iterdir():
        if item.is_dir():
            commons.append(item.name)

    for item in common_base.glob('*.py'):
        commons.append(item.stem)

    return list(sorted(filter(lambda x: not x.startswith('_'), commons)))


def is_common_package(name: str, commons: list[str]) -> str | None:
    split = name.split('.')

    if len(split) < 2:
        return None

    if split[0] != 'common':
        return None

    if split[1] not in commons:
        return None

    return split[1]


def find_common_deps_in_file(target_file, common_base, commons, common_imports):
    lg.debug(f'Processing {target_file}')
    file_imports = set()
    target_code = target_file.read_text(encoding='utf-8')
    target_ast = ast.parse(target_code)

    for stmt in target_ast.body:
        if isinstance(stmt, ast.Import):
            for name in sorted(stmt.names, key=lambda x: x.name):
                lg.debug(f'Found import "{name.name}" in {target_file}')

                if package := is_common_package(name.name, commons):
                    lg.debug(f'Adding "{package}" to file imports')
                    file_imports.add(package)

        if isinstance(stmt, ast.ImportFrom):
            lg.debug(f'Found import from "{stmt.module}" in {target_file}')

            if stmt.module:
                if package := is_common_package(stmt.module, commons):
                    lg.debug(f'Adding "{package}" to file imports')
                    file_imports.add(package)

    lg.debug(f'File imports: {target_file}: {file_imports}')
    new_imports = set(file_imports) - common_imports
    common_imports.update(new_imports)
    lg.debug(f'Common imports: {common_imports}')
    lg.debug(f'New imports: {new_imports}')

    for file_import in sorted(list(new_imports)):
        dep_file = common_base / file_import

        if not dep_file.is_dir():
            dep_file = dep_file.with_suffix('.py')
            lg.debug(f'Processing nested file {dep_file}')
            find_common_deps_in_file(dep_file, common_base, commons, common_imports)
        else:
            lg.debug(f'Processing nested directory {dep_file}')
            find_common_deps(dep_file, common_base, commons, common_imports)


def find_common_deps(target, common_base, commons, common_imports):
    target_files = sorted(target.glob('**/*.py'), key=lambda x: str(x))
    lg.debug(f'For target {target} files are: {target_files}')

    for target_file in target_files:
        find_common_deps_in_file(target_file, common_base, commons, common_imports)

    return common_imports



from common.aoss import create_client, INDEX_NAME


def handler(event, context):
    try:
        match event:
            case 'search':
                return search_aoss()
            case _:
                return 'Warning: Invalid event'

    except Exception as e:
        return f'Error: {str(e)}'


def search_aoss():
    client = create_client()
    query = {'query': {'match_all': {}}}
    return client.search(index=INDEX_NAME, body=query)



import sys
import os
from pathlib import Path


EXAMPLE_ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(EXAMPLE_ROOT.absolute()))


if 'AWS_PROFILE' not in os.environ:
    os.environ['AWS_PROFILE'] = 'easysam-a'


from common.myobject.db import ItemModel  # noqa: E402
from common.myobject.models import Item, NestedItem  # noqa: E402


def test_item_crud_flow():
    foo_value = 'Foo#1'
    bar_value = 'Bar#1'
    nested_initial = NestedItem(NestedFoo='nested#1')

    created = ItemModel.put(Item(Foo=foo_value, Bar=bar_value, Baz='baz', Nested=nested_initial))
    assert created.Bar == bar_value
    assert created.Nested.NestedFoo == 'nested#1'

    persisted = ItemModel.get(foo=foo_value, bar=bar_value)
    assert persisted.Baz == 'baz'
    assert persisted.Nested.NestedFoo == 'nested#1'

    updated = ItemModel.update(
        ItemModel.UpdateDTO(Baz='updated', Nested=NestedItem(NestedFoo='nested#2')),
        foo=foo_value,
        bar=bar_value,
    )
    assert updated.Baz == 'updated'
    assert updated.Nested.NestedFoo == 'nested#2'

    listed = ItemModel.list(foo=foo_value)
    assert len(listed) == 1
    assert listed[0].Baz == 'updated'
    assert listed[0].Nested.NestedFoo == 'nested#2'

    scanned = ItemModel.scan()
    assert len(scanned) == 1
    assert scanned[0].Foo == foo_value
    assert scanned[0].Nested.NestedFoo == 'nested#2'

    final = ItemModel.save(
        Item(Foo=foo_value, Bar=bar_value, Baz='saved', Nested=NestedItem(NestedFoo='nested#3')),
        original=updated,
    )
    assert final.Baz == 'saved'
    assert final.Nested.NestedFoo == 'nested#3'

    ItemModel.delete(foo=foo_value, bar=bar_value)
    assert ItemModel.scan() == []



import logging as lg
import json
import time
import shutil
from pathlib import Path
import subprocess

from benedict import benedict
from rich.live import Live
from rich.spinner import Spinner

from easysam.generate import generate
from easysam.commondep import commondep
import easysam.utils as u

SAM_CLI_VERSION = '1.138.0'
PIP_VERSION = '25.1.1'


def deploy(cliparams: dict, directory: Path, deploy_ctx: benedict):
    '''
    Deploy a SAM template to AWS.

    Args:
        cliparams: The CLI parameters.
        directory: The directory containing the SAM template.
        deploy_ctx: The deployment context.
    '''

    resources, errors = generate(cliparams, directory, [], deploy_ctx)

    if errors:
        lg.error(f'There were {len(errors)} errors:')

        for error in errors:
            lg.error(error)

        raise UserWarning('There were errors - aborting deployment')

    lg.info(f'Deploying SAM template from {directory}')
    check_pip_version(cliparams)
    check_sam_cli_version(cliparams)
    remove_common_dependencies(directory)
    copy_common_dependencies(directory, resources)

    # Building the application from the SAM template
    sam_build(cliparams, directory)

    # Deploying the application to AWS
    sam_deploy(cliparams, directory, deploy_ctx, resources)

    if not cliparams.get('no_cleanup'):
        remove_common_dependencies(directory)


def delete(cliparams, environment):
    lg.info(f'Deleting SAM template from {environment}')
    force = cliparams.get('force')
    await_deletion = cliparams.get('await_deletion')
    cf = u.get_aws_client('cloudformation', cliparams)
    mode = 'FORCE_DELETE_STACK' if force else 'STANDARD'
    cf.delete_stack(StackName=environment, DeletionMode=mode)  # type: ignore

    if await_deletion:
        lg.info(f'Awaiting deletion of {environment}')
        stack_status = 'UNKNOWN'

        with Live(Spinner('aesthetic', 'Checking stack status...'), transient=True):
            while stack_status != 'DELETE_COMPLETE':
                lg.debug(f'Stack {environment} is {stack_status}')

                if stack_status != 'UNKNOWN':
                    time.sleep(10)

                try:
                    stacks = cf.describe_stacks(StackName=environment).get('Stacks')
                except Exception as e:
                    lg.debug(f'Error describing stack {environment}: {e}')
                    break

                if not stacks:
                    lg.info(f'Stack {environment} deleted')
                    break

                stack_status = stacks[0]['StackStatus']

                if stack_status not in ['DELETE_COMPLETE', 'DELETE_IN_PROGRESS']:
                    raise UserWarning(f'Stack {environment} is {stack_status}')

    lg.info(f'Stack {environment} deleted')


def check_pip_version(cliparams):
    lg.info('Checking pip version')

    try:
        lg.debug(f'Running command: {" ".join(["pip", "--version"])}')
        pip_version = subprocess.check_output(['pip', '--version']).decode('utf-8')

        if pip_version < PIP_VERSION:
            raise UserWarning(f'pip version must be {PIP_VERSION} or higher')

    except Exception as e:
        raise UserWarning('pip not found') from e


def check_sam_cli_version(cliparams):
    lg.info('Checking SAM CLI version')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')
    sam_params.append('--version')

    try:
        lg.debug(f'Running command: {" ".join(sam_params)}')
        sam_version = subprocess.check_output(sam_params).decode('utf-8')
        lg.debug(f'SAM CLI version: {sam_version}')

        if sam_version < SAM_CLI_VERSION:
            raise UserWarning(f'SAM CLI version must be {SAM_CLI_VERSION} or higher')

    except Exception as e:
        raise UserWarning(f'SAM CLI not found. Error: {e}') from e


def sam_build(cliparams, directory):
    lg.info(f'Building SAM template from {directory}')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')
    sam_params.append('build')

    if cliparams.get('verbose'):
        sam_params.append('--debug')

    try:
        lg.debug(f'Running command: {" ".join(sam_params)}')
        subprocess.run(sam_params, cwd=directory.resolve(), text=True, check=True)
        lg.info('Successfully built SAM template')

    except subprocess.CalledProcessError as e:
        lg.error(f'Failed to build SAM template: {e}')
        raise UserWarning('Failed to build SAM template') from e


def sam_deploy(cliparams, directory, deploy_ctx, resources):
    lg.info(f'Deploying SAM template from {directory} to\n{json.dumps(deploy_ctx, indent=4)}')
    sam_tool = cliparams['sam_tool']
    sam_params = sam_tool.split(' ')

    aws_stack = deploy_ctx['environment']

    if not aws_stack:
        raise UserWarning('No AWS stack found in deploy context')

    sam_params.extend([
        'deploy',
        '--parameter-overrides', f'ParameterKey=Stage,ParameterValue={aws_stack}',
        '--stack-name', aws_stack,
        '--no-fail-on-empty-changeset',
        '--no-confirm-changeset',
        '--resolve-s3',
        '--capabilities', 'CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM'
    ])

    region = deploy_ctx.get('target_region')

    if region:
        sam_params.extend(['--region', region])

    aws_tags = list(cliparams.get('tag', []))

    if 'tags' in resources:
        aws_tags.extend(
            f'{name}={value}'
            for name, value in resources['tags'].items()
        )

    aws_tag_string = ' '.join(aws_tags)
    lg.info(f'AWS tag string: {aws_tag_string}')
    sam_params.extend(['--tags', aws_tag_string])

    if cliparams.get('verbose'):
        sam_params.append('--debug')

    if cliparams['dry_run']:
        lg.info(f'Would run: {" ".join(sam_params)}')
        return

    if cliparams.get('aws_profile'):
        lg.info(f'Using AWS profile: {cliparams["aws_profile"]}')
        sam_params.extend(['--profile', cliparams['aws_profile']])

    try:
        lg.debug(f'Running command: {" ".join(sam_params)}')
        subprocess.run(sam_params, cwd=directory.resolve(), text=True, check=True)
        lg.info('Successfully deployed SAM template')

    except subprocess.CalledProcessError as e:
        lg.error(f'Failed to deploy SAM template: {e}')
        raise UserWarning('Failed to deploy SAM template') from e


def common_dep_dir(directory):
    return Path(directory, 'common')


def remove_common_dependencies(directory):
    lg.info(f'Removing common dependencies from {directory}')
    backend = Path(directory, 'backend')

    for common_dep in backend.glob('**/common'):
        lg.debug(f'Removing {common_dep}')
        shutil.rmtree(common_dep)


def copy_common_dependencies(directory, resources):
    lg.info('Looking for common dependencies')
    common = common_dep_dir(directory)

    if not common.exists():
        lg.info('No common dependencies found')
        return

    lg.info(f'Copying common dependencies to {directory}')

    if 'functions' not in resources:
        lg.warning('No functions found in resources')
        return

    for lambda_name, lambda_function in resources['functions'].items():
        lambda_path = Path(directory, lambda_function['uri'])
        lambda_common_path = Path(lambda_path, 'common')
        lambda_common_path.mkdir(parents=True, exist_ok=True)
        deps = commondep(common, lambda_path)

        lg.info(f'Lambda {lambda_name} has {len(deps)} common dependencies')
        lg.debug(f'Dependencies: {" ".join(deps)}')

        for dep in deps:
            dep_path = Path(common, dep)

            if dep_path.is_dir():
                lg.debug(f'Copying {dep_path} directory to {lambda_common_path}')
                lambda_common_dep_path = Path(lambda_common_path, dep_path.name)
                shutil.copytree(dep_path, lambda_common_dep_path)
            else:
                dep_filepath = dep_path.with_suffix('.py')
                lg.debug(f'Copying {dep_filepath} file to {lambda_common_path}')
                lambda_common_filepath = Path(lambda_common_path, dep_filepath.name)
                shutil.copy(dep_filepath, lambda_common_filepath)



from pathlib import Path
import logging as lg


UTILS_PY = '''\
# Common code

def my_common_function():
    return 'Hello, world!'
'''

RESOURCES_YAML = '''\
prefix: MyApp

tags:
  mytag: myvalue

import:
  - backend
'''

LAMBDA_EASY_SAM = '''\
lambda:
  name: myfunction
  resources:
    tables:
    - MyItem
  integration:
    path: /items
    open: true
    greedy: false
'''

INDEX_PY = '''\
# Lambda function code

import json

import common.utils as u


def handler(event, context):
    data = u.my_common_function()

    return {
        'statusCode': 200,
        'headers': {
            'Access-Control-Allow-Headers': '*',
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': '*'
        },
        'body': json.dumps({'data': data})
    }
'''


DATABASE_EASY_SAM = '''\
tables:
  MyItem:
    attributes:
    - hash: true
      name: ItemID
'''

ROOT_GITIGNORE = '''\
build
template.yml
template.yaml
swagger.yaml
prismarine_client.py
.aws-sam
__pycache__/
'''

FUNCTION_GITIGNORE = '''\
*.py[oc]
**/common/
prismarine_client.py
'''

REQUIREMENTS_TXT = '''\
boto3
'''

REQUIREMENTS_TXT_PRISMARINE = '''\
boto3
prismarine
'''

RESOURCES_YAML_PRISMARINE = '''\
prefix: MyApp

import:
  - backend

prismarine:
  default-base: common
  access-module: common.dynamo_access
  tables:
    - package: myobject
'''

UTILS_PY_PRISMARINE = '''\
import os


def get_env():
    return os.environ.get('ENV', 'dev')
'''

DYNAMO_ACCESS_PY = '''\
import boto3

from prismarine.runtime.dynamo_access import DynamoAccess
from common.utils import get_env


class MyDynamoAccess(DynamoAccess):
    def get_resource(self):
        return boto3.resource('dynamodb')

    def get_table(self, full_model_name: str):
        env = get_env()
        return self.get_resource().Table(f'{full_model_name}-{env}')


dynamoaccess = MyDynamoAccess()


def get_dynamo_access():
    return dynamoaccess
'''

MODELS_PY = '''\
from typing import TypedDict, NotRequired
from prismarine.runtime import Cluster


c = Cluster('MyApp')


@c.model(PK='Foo', SK='Bar', trigger='itemlogger')
class Item(TypedDict):
    Foo: str
    Bar: str
    Baz: NotRequired[str]
'''

DB_PY = '''\
import prismarine_client as pc


class ItemModel(pc.ItemModel):
    pass
'''

TRIGGER_LAMBDA_EASY_SAM = '''\
lambda:
  name: itemlogger
  resources:
    tables:
    - Item
'''

TRIGGER_INDEX_PY = '''\
import json
import logging as lg

lg.basicConfig(level=lg.INFO)
logger = lg.getLogger()


def handler(event, context):
    """
    DynamoDB Stream trigger handler that logs Item changes.
    This is a simple example demonstrating trigger functionality with Prismarine models.
    """
    logger.info(f'Received event: {json.dumps(event)}')

    for record in event.get('Records', []):
        event_name = record.get('eventName')
        logger.info(f'Event: {event_name}')

        if event_name == 'INSERT':
            new_image = record['dynamodb'].get('NewImage', {})
            logger.info(f'New item created: {new_image}')

        elif event_name == 'MODIFY':
            new_image = record['dynamodb'].get('NewImage', {})
            old_image = record['dynamodb'].get('OldImage', {})
            logger.info(f'Item modified from {old_image} to {new_image}')

        elif event_name == 'REMOVE':
            old_image = record['dynamodb'].get('OldImage', {})
            logger.info(f'Item removed: {old_image}')

    return {
        'statusCode': 200,
        'body': json.dumps({'message': f'Processed {len(event.get("Records", []))} records'})
    }
'''


def init(cliparams, prismarine=False):
    app_dir = Path('.')
    pyproject_path = app_dir / 'pyproject.toml'
    if not pyproject_path.exists():
        raise UserWarning('pyproject.toml not found. Please run "uv init" first to initialize the project.')
    lg.info(f'Initializing app in current directory {"with Prismarine support" if prismarine else ""}')
    lg.debug(f'Creating resources EasySAM file {app_dir / "resources.yaml"}')
    resources_content = RESOURCES_YAML_PRISMARINE if prismarine else RESOURCES_YAML
    app_dir.joinpath('resources.yaml').write_text(resources_content)
    lg.debug(f'Creating root .gitignore file {app_dir / ".gitignore"}')
    gitignore_path = app_dir / '.gitignore'
    easysam_entries = set(ROOT_GITIGNORE.strip().split('\n'))
    if gitignore_path.exists():
        existing_entries = set(gitignore_path.read_text().strip().split('\n'))
        merged_entries = existing_entries | easysam_entries
        gitignore_path.write_text('\n'.join(sorted(merged_entries, key=str.lower)) + '\n')
    else:
        gitignore_path.write_text(ROOT_GITIGNORE)

    common_dir = app_dir / 'common'
    common_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating common directory {common_dir}')
    utils_content = UTILS_PY_PRISMARINE if prismarine else UTILS_PY
    common_dir.joinpath('utils.py').write_text(utils_content)

    if prismarine:
        lg.debug(f'Creating dynamo_access.py file {common_dir / "dynamo_access.py"}')
        common_dir.joinpath('dynamo_access.py').write_text(DYNAMO_ACCESS_PY)
        myobject_dir = common_dir / 'myobject'
        lg.debug(f'Creating myobject directory {myobject_dir}')
        myobject_dir.mkdir(parents=True, exist_ok=True)
        lg.debug(f'Creating models.py file {myobject_dir / "models.py"}')
        myobject_dir.joinpath('models.py').write_text(MODELS_PY)
        lg.debug(f'Creating db.py file {myobject_dir / "db.py"}')
        myobject_dir.joinpath('db.py').write_text(DB_PY)

    backend_dir = app_dir / 'backend'
    lg.debug(f'Creating backend directory {backend_dir}')
    backend_dir.mkdir(parents=True, exist_ok=True)

    if not prismarine:
        lg.debug(f'Creating database directory {backend_dir}')
        database_dir = backend_dir / 'database'
        database_dir.mkdir(parents=True, exist_ok=True)
        lg.debug(f'Creating database EasySAM file {database_dir / "easysam.yaml"}')
        database_dir.joinpath('easysam.yaml').write_text(DATABASE_EASY_SAM)

    function_dir = backend_dir / 'function'
    lg.debug(f'Creating function directory {function_dir}')
    function_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating function .gitignore file {function_dir / ".gitignore"}')
    function_dir.joinpath('.gitignore').write_text(FUNCTION_GITIGNORE)

    if prismarine:
        lambda_dir = function_dir / 'itemlogger'
        lg.debug(f'Creating lambda directory {lambda_dir}')
        lambda_dir.mkdir(parents=True, exist_ok=True)
        lg.debug(f'Creating lambda EasySAM file {lambda_dir / "easysam.yaml"}')
        lambda_dir.joinpath('easysam.yaml').write_text(TRIGGER_LAMBDA_EASY_SAM)
        lg.debug(f'Creating lambda index file {lambda_dir / "index.py"}')
        lambda_dir.joinpath('index.py').write_text(TRIGGER_INDEX_PY)
    else:
        lambda_dir = function_dir / 'myfunction'
        lg.debug(f'Creating lambda directory {lambda_dir}')
        lambda_dir.mkdir(parents=True, exist_ok=True)
        lg.debug(f'Creating lambda EasySAM file {lambda_dir / "easysam.yaml"}')
        lambda_dir.joinpath('easysam.yaml').write_text(LAMBDA_EASY_SAM)
        lg.debug(f'Creating lambda index file {lambda_dir / "index.py"}')
        lambda_dir.joinpath('index.py').write_text(INDEX_PY)

    third_party_dir = app_dir / 'thirdparty'
    lg.debug(f'Creating third party directory {third_party_dir}')
    third_party_dir.mkdir(parents=True, exist_ok=True)
    lg.debug(f'Creating third party requirements.txt file {third_party_dir / "requirements.txt"}')
    requirements_content = REQUIREMENTS_TXT_PRISMARINE if prismarine else REQUIREMENTS_TXT
    third_party_dir.joinpath('requirements.txt').write_text(requirements_content)



import boto3
import logging as lg
from benedict import benedict


def scan(resources_data: benedict, errors: list[str], aws_profile: str = None):
    # Forward compatibility with future resources
    if '__future__' not in resources_data:
        return

    cloud = resources_data.get_dict('cloud', {})

    lg.info('Scanning cloud for resources')

    session_kwargs = {}
    if aws_profile:
        session_kwargs['profile_name'] = aws_profile

    session = boto3.Session(**session_kwargs)
    ec2 = session.client('ec2')

    vpcs = ec2.describe_vpcs()['Vpcs']

    if not vpcs:
        errors.append('Error: No VPCs found')
        return

    if len(vpcs) > 1:
        errors.append('Error: Not implemented for multiple VPCs')
        return

    vpc = vpcs[0]
    vpc_id = vpc['VpcId']
    cloud['vpc_id'] = vpc_id
    cloud['vpc_cidr'] = vpc['CidrBlock']

    lg.info(f'Found VPC {vpc_id} with CIDR {cloud["vpc_cidr"]}')

    subnets = ec2.describe_subnets(Filters=[{'Name': 'vpc-id', 'Values': [vpc_id]}])[
        'Subnets'
    ]
    subnet_ids = [subnet['SubnetId'] for subnet in subnets]

    lg.info(f'Found {len(subnet_ids)} subnets in VPC {vpc_id}')
    lg.debug(f'Subnet IDs: {subnet_ids}')

    cloud['subnet_ids'] = subnet_ids
    resources_data['cloud'] = cloud



from pathlib import Path
import logging as lg
import sys
from importlib.metadata import version
import traceback
from argparse import ArgumentParser

from benedict import benedict
import click

from easysam.generate import generate
from easysam.deploy import deploy, delete
from easysam.deploy import remove_common_dependencies
from easysam.init import init

from easysam.inspect import inspect


@click.group(help='EasySAM is a tool for generating SAM templates from simple YAML files')
@click.version_option(version('easysam'))
@click.pass_context
@click.option(
    '--aws-profile', type=str, help='AWS profile to use'
)
@click.option(
    '--context-file', type=click.Path(exists=True),
    help='A YAML file containing additional context for the resources.yaml file. '
         'For example, overrides for resource properties.'
)
@click.option(
    '--target-region', type=str, help='A region to use for generation'
)
@click.option(
    '--environment', type=str, help='An environment (AWS stack) to use in generation',
    default='dev'
)
@click.option(
    '--verbose', is_flag=True
)
def easysam(ctx, verbose, aws_profile, context_file, target_region, environment):
    ctx.obj = {
        'verbose': verbose,
        'aws_profile': aws_profile,
        'deploy_ctx': {
            'target_region': target_region,
            'environment': environment
        }
    }

    if context_file:
        ctx.obj['deploy_ctx'] = benedict.from_yaml(Path(context_file))
        lg.info(f'Loaded context from {context_file}')

    lg.basicConfig(level=lg.DEBUG if verbose else lg.INFO)
    lg.debug(f'Verbose: {verbose}')


@easysam.command(name='generate', help='Generate a SAM template from a directory')
@click.pass_obj
@click.option(
    '--path', multiple=True, help='A additional Python path to use for generation'
)
@click.argument('directory', type=click.Path(exists=True))
def generate_cmd(obj, directory, path):
    directory = Path(directory)
    pypath = [Path(p) for p in path]
    deploy_ctx = obj.get('deploy_ctx')
    resources_data, errors = generate(obj, directory, pypath, deploy_ctx)

    if errors:
        for error in errors:
            lg.error(error)

        if len(errors) > 1:
            lg.error(f'There were {len(errors)} errors:')
        else:
            lg.error('There was an error')

        sys.exit(1)

    else:
        click.echo(resources_data.to_yaml())
        lg.info('Resources generated successfully')
        sys.exit(0)


@easysam.command(name='deploy', help='Deploy the application to an AWS environment')
@click.pass_obj
@click.option(
    '--tag', type=str, multiple=True, help='AWS Tags'
)
@click.option(
    '--dry-run', is_flag=True, help='Dry run the deployment'
)
@click.option(
    '--sam-tool', type=str, help='Path to the SAM CLI', default='uv run sam'
)
@click.option(
    '--no-cleanup', is_flag=True, help='Do not clean the directory before deploying'
)
@click.option(
    '--override-main-template',
    type=click.Path(exists=True, path_type=Path),
    help='Override the main template',
)
@click.argument('directory', type=click.Path(exists=True, path_type=Path))
def deploy_cmd(obj, directory, **kwargs):
    obj.update(kwargs)  # noqa: F821
    deploy_ctx = obj.get('deploy_ctx')
    deploy(obj, directory, deploy_ctx)


@easysam.command(name='delete', help='Delete the environment from AWS')
@click.pass_obj
@click.option(
    '--force', is_flag=True, help='Force delete the environment'
)
@click.option(
    '--await', 'await_deletion', is_flag=True, help='Await the deletion to complete'
)
def delete_cmd(obj, **kwargs):
    obj.update(kwargs)  # noqa: F821
    environment = obj.get('deploy_ctx').get('environment')
    delete(obj, environment)


@easysam.command(name='cleanup', help='Remove common dependencies from the directory')
@click.pass_obj
@click.argument('directory', type=click.Path(exists=True))
def cleanup_cmd(obj, directory):
    remove_common_dependencies(directory)


@easysam.command(name='init', help='Initialize a new application in the current directory (requires uv init to be run first)')
@click.pass_obj
@click.option('--prismarine', is_flag=True, help='Scaffold a minimal application with Prismarine support')
def init_cmd(obj, prismarine):
    init(obj, prismarine=prismarine)


def main():
    try:
        easysam.add_command(inspect)
        easysam()

    except UserWarning as e:
        parser = ArgumentParser()
        parser.add_argument('--verbose', action='store_true')
        args, _ = parser.parse_known_args()

        if args.verbose:
            lg.error(traceback.format_exc())

        lg.error(e)
        sys.exit(1)

    except Exception as e:
        lg.error(traceback.format_exc())
        lg.error(f'An unexpected error occurred: {e}')
        sys.exit(2)


if __name__ == '__main__':
    main()



import logging as lg
from pathlib import Path
from typing import Any

from benedict import benedict
import yaml

import prismarine.prisma_common as prisma_common
import prismarine.prisma_easysam as prisma_easysam

from easysam.validate_schema import validate as validate_schema
from easysam.definitions import FatalError


IMPORT_FILE = 'easysam.yaml'

SUPPORTED_SECTIONS = [
    'tables',
    'paths',
    'functions',
    'buckets',
    'authorizers',
    'prismarine',
    'import',
    'lambda',
    'search',
    'mqtt',
]


STREAM_INTERVAL_SECONDS = 300


def resources(
    resources_dir: Path,
    pypath: list[Path],
    deploy_ctx: dict[str, str],
    errors: list[str]
) -> benedict:
    '''
    Load the resources from the resources.yaml file.

    Args:
        resources_dir: The directory containing the resources.yaml file.
        pypath: The additional Python path to use.
        deploy_ctx: The deployment context dictionary.
        errors: The list of errors.

    Returns:
        A dictionary containing the resources.
    '''

    resources = Path(resources_dir, 'resources.yaml')

    try:
        yaml.SafeLoader.add_constructor('!Conditional', conditional_constructor)
        raw_resources_data = benedict(yaml.safe_load(Path(resources).read_text(encoding='utf-8')))
    except Exception as e:
        errors.append(f'Error loading resources file {resources}: {e}')
        return benedict()

    lg.info('Resolving conditional resources')
    lg.debug(f'Deployment context: {deploy_ctx}')
    resources_data = resolve_conditionals(raw_resources_data, deploy_ctx, errors)
    lg.debug('Resources data after resolving conditionals:')
    lg.debug(resources_data.to_yaml())

    lg.info('Applying overrides')
    apply_overrides(resources_data, deploy_ctx)

    lg.info('Processing resources')
    pypath = [resources_dir] + list(pypath)
    preprocess_resources(resources_data, resources_dir, pypath, errors)

    lg.info('Validating resources')
    validate_schema(resources_dir, resources_data, errors)

    lg.info('Adding internal values')
    check_lambda_layer(resources_dir, resources_data)

    return resources_data


def prismarine_dynamo_tables(
    prefix: str,
    base: str,
    package: str,
    resources_dir: Path,
    pypath: list[Path],
    errors: list[str]
):
    lg.debug(f'Generating prismarine dynamo tables for {prefix}')

    try:
        prisma_common.set_path(pypath)
        base_dir = Path(resources_dir, base).resolve()
        cluster = prisma_common.get_cluster(base_dir, package)
        # TODO Amend prismarine to return errors
        return prisma_easysam.build_dynamo_tables(prefix, cluster)

    except Exception as e:
        lg.error(f'Error generating dynamo tables for {prefix}: {e}')
        errors.append(f'Error generating prismarine dynamo tables for {base_dir}')
        return None


def preprocess_prismarine(
    resources_data: dict,
    resources_dir: Path,
    pypath: list[Path],
    errors: list[str]
):
    prefix = resources_data['prefix']
    prisma = resources_data['prismarine']
    prisma_base = prisma.get('default-base')
    prisma_tables = prisma['tables'] or []

    for prisma_integration in prisma_tables:
        base = prisma_integration.get('base') or prisma_base
        package = prisma_integration.get('package')

        if not base:
            errors.append(f'No base found for {package}')
            continue

        if not package:
            errors.append(f'No package found for {base}')
            continue

        tables = prismarine_dynamo_tables(
            prefix, base, package, resources_dir, pypath, errors
        )

        if not tables:
            lg.warning(f'No valid tables found for {package}, continuing')
            continue

        if 'tables' not in resources_data:
            resources_data['tables'] = {}

        resources_data['tables'].update(tables)


def preprocess_lambda(
    resources_data: dict,
    resources_dir: Path,
    lambda_def: dict,
    entry_path: Path,
    entry_dir: Path,
    errors: list[str]
):
    if 'functions' not in resources_data:
        resources_data['functions'] = {}

    lambda_name = lambda_def.get('name')

    if not lambda_name:
        errors.append(f'Import file {entry_path} contains no lambda name')
        return

    if lambda_name in resources_data['functions']:
        errors.append(
            f'Import file {entry_path} contains duplicate lambda name {lambda_name}'
        )
        return

    lambda_resources = lambda_def.get('resources', {})

    if 'uri' not in lambda_resources:
        lg.debug(f'Adding uri to lambda {lambda_name}')
        lambda_resources['uri'] = Path(entry_dir).relative_to(resources_dir).as_posix()

    lg.debug(f'Adding lambda {lambda_name} to resources')
    resources_data['functions'][lambda_name] = lambda_resources
    integration = lambda_def.get('integration', {})

    if integration:
        path = integration.get('path')

        if not path:
            errors.append(f'Import file {entry_path} contains no path name')
            return

        if 'paths' not in resources_data:
            resources_data['paths'] = {}

        if path in resources_data['paths']:
            errors.append(f'Import file {entry_path} contains duplicate path {path}')
            return

        integration['function'] = lambda_name
        integration['integration'] = 'lambda'
        del integration['path']
        lg.debug(f'Adding path {path} to resources')
        resources_data['paths'][path] = integration


def preprocess_tables(
    resources_data: dict,
    table_def: dict,
    entry_path: Path,
    errors: list[str]
):
    if 'tables' not in resources_data:
        resources_data['tables'] = {}

    for table_name, table_data in table_def.items():
        if table_name in resources_data['tables']:
            errors.append(f'Import file {entry_path} contains duplicate table {table_name}')
            continue

        lg.debug(f'Adding table {table_name} to resources')
        resources_data['tables'][table_name] = table_data


def preprocess_file(
    resources_data: dict,
    resources_dir: Path,
    entry_path: Path,
    errors: list[str]
):
    lg.info(f'Processing import file {entry_path}')
    try:
        entry_dir = entry_path.parent
        entry_data = yaml.safe_load(entry_path.read_text(encoding='utf-8'))
    except Exception as e:
        errors.append(f'Error loading import file {entry_path}: {e}')
        return

    if not all(key in ['lambda', 'import', 'tables'] for key in entry_data.keys()):
        errors.append(f'Import file {entry_path} contains unexpected sections')
        return

    if lambda_def := entry_data.get('lambda'):
        preprocess_lambda(
            resources_data, resources_dir, lambda_def, entry_path, entry_dir, errors
        )

    if tables_def := entry_data.get('tables'):
        preprocess_tables(resources_data, tables_def, entry_path, errors)

    if local_import_def := entry_data.get('import'):
        for import_file in local_import_def:
            import_path = Path(entry_dir, import_file)
            preprocess_file(resources_data, resources_dir, import_path, errors)


def preprocess_imports(resources_data: dict, resources_dir: Path, errors: list[str]):
    for import_dir_str in resources_data.get('import', []):
        import_dir = Path(resources_dir, import_dir_str)
        lg.info(f'Processing import directory {import_dir}')

        if not import_dir.exists():
            errors.append(f'Import directory {import_dir} not found')
            continue

        for entry_path in import_dir.glob(f'**/{IMPORT_FILE}'):
            preprocess_file(resources_data, resources_dir, entry_path, errors)


def process_default_functions(resources_data: dict, errors: list[str]):
    def transform_lambda_poll(poll: str | dict):
        if isinstance(poll, str):
            return {'name': poll}
        else:
            return poll

    if 'functions' in resources_data:
        for name, function in resources_data['functions'].items():
            lg.debug(f'Processing function {name}')

            if polls := function.get('polls'):
                function['polls'] = [transform_lambda_poll(poll) for poll in polls]

            if 'searches' in function:
                if not function['searches']:
                    function['searches'] = ['searchable']


def process_default_streams(resources_data: dict, errors: list[str]):
    new_streams = {}

    for stream_name, stream in resources_data.get('streams', {}).items():
        if 'buckets' in stream and 'bucketname' in stream:
            errors.append(f'Stream {stream} cannot have both buckets and bucketname')
            continue

        if 'bucketname' in stream:
            stream['buckets'] = {
                'private': {
                    'bucketname': stream['bucketname'],
                    'bucketprefix': stream.get('bucketprefix', ''),
                    'intervalinseconds': stream.get('intervalinseconds')
                }
            }

            del stream['bucketname']

            if 'bucketprefix' in stream:
                del stream['bucketprefix']

            if 'intervalinseconds' in stream:
                del stream['intervalinseconds']

        for bucket in stream.get('buckets', {}).values():
            if 'bucketprefix' not in bucket:
                bucket['bucketprefix'] = ''

            if 'intervalinseconds' not in bucket:
                bucket['intervalinseconds'] = STREAM_INTERVAL_SECONDS

        new_streams[stream_name] = stream

    if new_streams:
        resources_data['streams'] = new_streams


def process_default_paths(resources_data: dict, errors: list[str]):
    for path in resources_data.get('paths', {}).values():
        if 'integration' not in path:
            path['integration'] = 'lambda'

        match path.get('integration'):
            case 'dynamo':
                path['action'] = path.get('action', 'GetItem')
            case 'sqs':
                path['method'] = path.get('method', 'post')
            case 'lambda':
                if 'greedy' not in path:
                    path['greedy'] = True


def process_default_tables(resources_data: dict, errors: list[str]):
    for table_name, table in resources_data.get('tables', {}).items():
        if trigger_config := table.get('trigger'):
            # If trigger is a string, convert to object
            if isinstance(trigger_config, str):
                trigger_config = {'function': trigger_config}
                table['trigger'] = trigger_config

            # Set default viewtype to 'new-and-old'
            if 'viewtype' not in trigger_config:
                trigger_config['viewtype'] = 'new-and-old'

            # Set default startingposition to 'latest'
            if 'startingposition' not in trigger_config:
                trigger_config['startingposition'] = 'latest'


def preprocess_defaults(resources_data: dict, errors: list[str]):
    process_default_searches(resources_data, errors)
    process_default_functions(resources_data, errors)
    process_default_streams(resources_data, errors)
    process_default_tables(resources_data, errors)
    process_default_paths(resources_data, errors)


def preprocess_resources(
    resources_data: dict,
    resources_dir: Path,
    pypath: list[Path],
    errors: list[str]
):
    def sort_dict(d):
        return dict(sorted(d.items(), key=lambda x: x[0]))

    if 'prismarine' in resources_data:
        preprocess_prismarine(resources_data, resources_dir, pypath, errors)

    if 'import' in resources_data:
        preprocess_imports(resources_data, resources_dir, errors)

    preprocess_defaults(resources_data, errors)

    for section in SUPPORTED_SECTIONS:
        if section in resources_data:
            if isinstance(resources_data[section], dict):
                resources_data[section] = sort_dict(resources_data[section])
            elif isinstance(resources_data[section], list):
                resources_data[section] = sorted(resources_data[section])

    resources_data = sort_dict(resources_data)


def check_lambda_layer(resources_dir: Path, resources_data: dict):
    thirdparty_dir = Path(resources_dir, 'thirdparty')
    resources_data['enable_lambda_layer'] = thirdparty_dir.exists()


class Conditional(yaml.YAMLObject):
    def __init__(self, key, environment='any', region='any'):
        self.key = key
        self.environment = environment
        self.region = region

    def __repr__(self):
        fields = [
            f'key={self.key}',
            f'environment={self.environment}',
            f'region={self.region}',
        ]

        return f'Conditional({", ".join(fields)})'


def conditional_constructor(loader, node):
    mapping = loader.construct_mapping(node)

    if 'key' not in mapping:
        raise ValueError('All !Conditional keys must have a key')

    attributes = {'key': mapping['key']}

    if 'environment' in mapping:
        attributes['environment'] = mapping['environment']

    if 'region' in mapping:
        attributes['region'] = mapping['region']

    return Conditional(**attributes)


def check_condition(
    condition: str,
    value: str,
    deploy_ctx: dict[str, str],
    errors: list[str]
):
    if value == 'any':
        return True

    context_value = deploy_ctx.get(condition)

    if context_value is None:
        errors.append(
            f'Fatal error: Condition "{condition}" not found in deployment context. '
            f'Unable to resolve conditional resources. Consider adding "--{condition}"'
        )

        raise FatalError(errors)

    negate = value.startswith('~')
    value = value.lstrip('~')

    lg.info(
        f'Checking condition "{condition}" '
        f'{value} (condition) == {context_value} (context) (negate={negate})'
    )

    include = value == context_value
    result = include if not negate else not include
    return result


def resolve_conditionals(
    resources_data: dict,
    deploy_ctx: dict[str, str],
    errors: list[str]
):
    resolved = benedict()

    for key, value in resources_data.items():
        if isinstance(value, dict):
            resolved_value = resolve_conditionals(value, deploy_ctx, errors)
        else:
            resolved_value = value

        if isinstance(key, Conditional):
            include = all([
                check_condition('environment', key.environment, deploy_ctx, errors),
                check_condition('target_region', key.region, deploy_ctx, errors),
            ])

            if include:
                resolved[key.key] = resolved_value
        else:
            resolved[key] = resolved_value

    return resolved


def apply_overrides(resources_data: dict, deploy_ctx: dict[str, Any]):
    if 'overrides' in deploy_ctx:
        for override_path, override_value in deploy_ctx['overrides'].items():
            key = override_path.replace('/', '.')
            lg.info(f'Applying override: {key} = {override_value}')
            resources_data[key] = override_value


def process_default_searches(resources_data: dict, errors: list[str]):
    lg.debug('Processing default searches')

    if 'search' in resources_data:
        lg.debug('Search is defined')

        if not resources_data['search']:
            lg.debug('Search is empty, adding searchable')
            resources_data['search'] = {
                'searchable': {}
            }



import os

import boto3
from opensearchpy import (
    OpenSearch, RequestsHttpConnection, AWSV4SignerAuth
)

INDEX_NAME = 'searchable-documents'


def create_client():
    service = 'aoss'
    region = os.getenv('REGION')
    collection = os.getenv('SEARCH_SEARCHABLE_COLLECTION_ID')
    host = f'{collection}.{region}.{service}.amazonaws.com'
    credentials = boto3.Session().get_credentials()
    auth = AWSV4SignerAuth(credentials, region, service)
    return OpenSearch(
        hosts=[{'host': host, 'port': 443}],
        http_auth=auth,
        use_ssl=True,
        verify_certs=True,
        connection_class=RequestsHttpConnection,
        pool_maxsize=20
    )



import json

import common.utils as u


def handler(event, context):
    data = u.my_common_function()

    return {
        'statusCode': 200,
        'headers': {
            'Access-Control-Allow-Headers': '*',
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': '*'
        },
        'body': json.dumps({'data': data})
    }



import json


def handler(event, context):
    return {
        'statusCode': 200,
        'headers': {
            'Access-Control-Allow-Headers': '*',
            'Access-Control-Allow-Origin': '*',
            'Access-Control-Allow-Methods': '*'
        },
        'body': json.dumps({'data': 'Hello, World!'})
    }



# This file is generated by prismarine 1.5.2. Do not modify it.

from prismarine.runtime.dynamo_crud import (
    _query,
    _get_item,
    _update,
    _put_item,
    _delete,
    _scan,
    _save,
    Model,
    without,
    DbNotFound
)

from typing import List, Literal  # noqa
from types import EllipsisType

from common.dynamo_access import get_dynamo_access

from typing import Any, Optional, get_args, get_origin
from pydantic import BaseModel, Field, create_model
from pydantic.fields import PydanticUndefined


def _is_optional_type(annotation: Any):
    origin = get_origin(annotation)
    if origin:
        return type(None) in get_args(annotation)

    return annotation is type(None)


def _make_optional(annotation: Any):
    if _is_optional_type(annotation):
        return annotation

    try:
        return annotation | None  # type: ignore[operator]
    except TypeError:
        return Optional[annotation]


def _build_pydantic_update_model(model_cls):
    field_definitions = {}
    for name, field in model_cls.model_fields.items():
        annotation = _make_optional(field.annotation or Any)
        field_is_required = (
            field.is_required() if callable(field.is_required) else field.is_required
        )
        alias = field.alias if field.alias and field.alias != name else None

        if field_is_required:
            default = None
        else:
            default = field.default if field.default is not PydanticUndefined else None

        default_factory = None if field_is_required else field.default_factory
        field_kwargs = {}
        if alias:
            field_kwargs['alias'] = alias

        if default_factory is not None:
            field_definitions[name] = (
                annotation,
                Field(default_factory=default_factory, **field_kwargs)
            )
        elif field_kwargs:
            field_definitions[name] = (
                annotation,
                Field(default=default, **field_kwargs)
            )
        else:
            field_definitions[name] = (annotation, default)

    return create_model(
        f'{model_cls.__name__}UpdateDTO',
        __base__=BaseModel,
        __module__=__name__,
        **field_definitions
    )


def _dump_model(instance):
    if isinstance(instance, BaseModel):
        return instance.model_dump(mode='python', by_alias=True)

    return instance


def _dump_update_model(instance):
    if isinstance(instance, BaseModel):
        return instance.model_dump(mode='python', by_alias=True, exclude_unset=True)

    return instance


def _load_model(model_cls, value):
    if isinstance(value, model_cls):
        return value

    return model_cls.model_validate(value)


def _load_models(model_cls, items):
    return [_load_model(model_cls, item) for item in items]


# Model classes imports
from common.myobject.models import Item  # noqa: F401, E402
from common.myobject.models import NestedItem  # noqa: F401, E402

# Create a DynamoDB resource
dynamo = get_dynamo_access()


class ItemModel(Model):
    table_name = 'PrismaPydanticItem'
    PK = 'Foo'
    SK = 'Bar'

    UpdateDTO = _build_pydantic_update_model(Item)

    @staticmethod
    def list(*, foo: str, **kwargs) -> List[Item]:
        items = _query(dynamo, ItemModel.table_name, {ItemModel.PK: foo}, **kwargs)
        return _load_models(Item, items)

    @staticmethod
    def get(
        *, bar: str, foo: str, default: Item | EllipsisType = ..., **kwargs
    ) -> Item:
        item = _get_item(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            default,
            **kwargs
        )
        if not item:
            raise DbNotFound(
                f'Item was not found in {ItemModel.table_name} with params foo={foo}, bar={bar}'
            )
        return _load_model(Item, item)

    @staticmethod
    def put(item: Item, **kwargs) -> Item:
        payload = _dump_model(item)
        _put_item(dynamo, ItemModel.table_name, payload, **kwargs)
        return _load_model(Item, payload)

    @staticmethod
    def update(
        item: UpdateDTO, *, foo: str, bar: str, default: Item | EllipsisType = ...
    ) -> Item:
        initial = ItemModel.get(bar=bar, foo=foo, default=default)
        update_data = _dump_update_model(item)
        _update(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            without(update_data, [ItemModel.PK, ItemModel.SK])
        )
        return _load_model(Item, {**_dump_model(initial), **update_data})

    @staticmethod
    def save(updated: Item, *, original: Item | None = None) -> Item:
        updated_payload = _dump_model(updated)
        original_payload = _dump_model(original) if original else None
        _save(
            dynamo,
            ItemModel.table_name,
            kv={
                ItemModel.PK: updated_payload[ItemModel.PK],
                ItemModel.SK: updated_payload[ItemModel.SK]
            },
            updated=updated_payload,
            original=original_payload
        )
        return _load_model(Item, updated_payload)

    @staticmethod
    def delete(*, bar: str, foo: str, **kwargs):
        _delete(
            dynamo,
            ItemModel.table_name,
            {ItemModel.PK: foo, ItemModel.SK: bar},
            **kwargs
        )

    @staticmethod
    def scan(**kwargs) -> List[Item]:
        items = _scan(dynamo, ItemModel.table_name, **kwargs)
        return _load_models(Item, items)
